#!/usr/bin/env python3

from dataclasses import dataclass
from .baseparser import ParseError, BaseParser


@dataclass
class Decision:
    var: str
    type: str
    value: list
    desc: str = ''


class DecisionParser(BaseParser):
    def __init__(self, spec):
        super(DecisionParser, self).__init__('')
        self.spec = spec
        self.decisions = {}

    @staticmethod
    def _is_syntax_start(ch):
        return ch == '{'

    @staticmethod
    def _is_id_token(s):
        if not BaseParser._is_id_start(s[0]):
            return False

        for ch in s:
            if not BaseParser._is_id(ch):
                return False

        return True

    @staticmethod
    def _is_type(s):
        return s == 'discrete'

    @staticmethod
    def _read_value(s):
        try:
            res = list(s)
        except ValueError:
            raise ParseError('Cannot handle value "{}"'.format(s))

        if len(s) == 0:
            raise ParseError('Cannot handle decision value "[]"')
        return res

    def _read_json_safe(self, obj, field):
        if field not in obj:
            raise ParseError('Cannot find "{}" in json'.format(field))
        return obj[field]

    def _check_type(self, var, fun, msg):
        if not fun(var):
            raise ParseError('Cannot handle {} "{}"'.format(msg, var))
        return var

    def read_decisions(self):
        """
        Read decisions from the JSON spec.
        :return: a dict of decisions
        """
        res = {}

        dec_spec = self.spec['decisions'] if 'decisions' in self.spec else []
        for d in dec_spec:
            desc = d['desc'] if 'desc' in d else 'Decision {}'.format(d['var'])

            var = self._check_type(self._read_json_safe(d, 'var'),
                                   DecisionParser._is_id_token, 'id')
            tp = self._check_type(self._read_json_safe(d, 'type'),
                                  DecisionParser._is_type, 'type')
            value = self._read_value(self._read_json_safe(d, 'value'))

            decision = Decision(var, tp, value, desc)
            res[var] = decision

        self.decisions = res
        return res

    def get_num_alt(self, dec):
        """
        Return the number of possible alternatives.
        For discrete type, it's the number of values.
        :param dec: variable ID of a decision
        :return: number
        """
        return len(self.decisions[dec].value)

    def gen_code(self, template, dec_id, i_alt):
        """
        Replace the placeholder variable in a template chunk.
        :param template: a chunk of code with only one placeholder
        :param dec_id: variable ID of the decision
        :param i_alt: which alternative
        :return: {string, string} replaced code and the value at this parameter
        """
        dec = self.decisions[dec_id]
        v = dec.value[i_alt]

        # assuming the placeholder var is always at the end
        # which is true given how we chop up the chunks
        length = 4 + len(dec_id)
        return template[:-length] + str(v), str(v)

    def parse_code(self, line):
        """
        Find placeholder variables in a line of code.
        :param line: a line of code
        :return: (variables, chunks) a list of found variables and a list of
                 code chunks, each containing one variable
        """
        code = []
        res = []
        j = 0
        self.i = 0
        self.line = line

        while not self._is_end():
            if self._is_syntax_start(self._peek_char()):
                token = self._read_while(lambda ch: ch == '{')
                if len(token) < 2:
                    continue
                val = self._read_while(self._is_id)
                if len(val) == 0:
                    continue
                token = self._read_while(lambda ch: ch == '}', max_len=2)
                if len(token) < 2:
                    continue

                # read succeeds
                if val not in set(self.decisions.keys()):
                    msg = 'Cannot find the matching variable "{}" in spec'.format(val)
                    raise ParseError(msg)

                code.append(line[j:self.i])
                res.append(val)
                j = self.i
            else:
                self._next_char()

        code.append(line[j:])
        return res, code



#!/usr/bin/env python3

import json
import os
import shutil
import csv
from textwrap import wrap
from dataclasses import dataclass, field
from typing import List

from .blockparser import BlockParser, ParseError
from .graphparser import GraphParser
from .graphanalyzer import GraphAnalyzer, InvalidGraphError
from .decisionparser import DecisionParser
from .lang import LangError, Lang
import boba.util as util


@dataclass
class Block:
    code: str = ''
    id: str = ''
    name: str = ''
    chunks: List = field(default_factory=lambda: [])


@dataclass
class History:
    path: int
    filename: str = ''
    decisions: List = field(default_factory=lambda: [])


exec_template = """\
#!/bin/sh

DIR="$( cd "$( dirname "${{BASH_SOURCE[0]}}" )" >/dev/null 2>&1 && pwd )"
prefix={}
suffix={}
num={}
i=1

while [ $i -le $num ]
do
  f="$DIR/$prefix$i$suffix"
  echo "{} $f"
  {} $f
  i=$(( i+1 ))
done
"""


class Parser:

    """ Parse everything """

    def __init__(self, f1, f2, out='.', lang=''):
        self.fn_script = f1
        self.fn_spec = f2
        self.out = os.path.join(out, 'multiverse/')
        try:
            self.lang = Lang(lang, f1)
        except LangError as e:
            self._throw(e.args[0])

        self.blocks = {}
        self.paths = []
        self.history = []

        # read spec
        with open(f2, 'rb') as f:
            self.spec = json.load(f)

        # initialize helper class
        self.dec_parser = DecisionParser(self.spec)

    def _throw(self, msg):
        util.print_fail(msg)
        exit(1)

    def _indent(self, msg):
        return '\n'.join(['\t' + l for l in msg.split('\n')])

    @staticmethod
    def _rename_dec(dec):
        return '_' + dec

    def _throw_parse_error(self, msg):
        prefix = 'In parsing file "{}":\n'.format(self.fn_script)
        self._throw(prefix + self._indent(msg))

    def _throw_spec_error(self, msg):
        prefix = 'In parsing file "{}":\n'.format(self.fn_spec)
        self._throw(prefix + self._indent(msg))

    def _add_block(self, block):
        # ignore empty block
        if block.id == '' and block.code == '':
            return

        # handle unnamed block
        if block.id == '':
            block.id = '_start' if len(self.blocks) == 0 else '_end'

        # check if id exists
        if block.id in self.blocks:
            err = 'Duplicated code block ID "{}"'.format(block.id)
            self._throw_parse_error(err)
        self.blocks[block.id] = block

    def _parse_blocks(self):
        try:
            self.dec_parser.read_decisions()
        except ParseError as e:
            self._throw_spec_error(e.args[0])

        with open(self.fn_script, 'r') as f:
            code = ''
            bl = Block()

            for line in f:
                if BlockParser.can_parse(line):
                    # end of the last block
                    bl.chunks.append(('', code))
                    code = ''
                    self._add_block(bl)

                    # parse the metadata and create a new block
                    try:
                        bp_id, bp_name = BlockParser(line).parse()
                        bl = Block('', bp_id, bp_name, [])
                    except ParseError as e:
                        self._throw_parse_error(e.args[0])
                else:
                    # match decision variables
                    try:
                        vs, codes = self.dec_parser.parse_code(line)
                        if len(vs):
                            # chop into more chunks
                            # combine first chunk with previous code
                            bl.chunks.append((vs[0], code + codes[0]))
                            for i in range(1, len(vs)):
                                bl.chunks.append((vs[i], codes[i]))

                            # remaining code after the last matched variable
                            code = codes[-1]
                        else:
                            code += line
                    except ParseError as e:
                        msg = 'At line "{}"\n\t{}'.format(line, e.args[0])
                        self._throw_parse_error(msg)

                    # TODO: fix legacy
                    bl.code += line

            # add the last block
            bl.chunks.append(('', code))
            self._add_block(bl)

    def _match_nodes(self, nodes):
        # nodes in spec and script should match
        for nd in nodes:
            if nd not in self.blocks:
                self._throw_spec_error('Cannot find matching node "{}" in script'.format(nd))

        for nd in self.blocks:
            # ignore special nodes inserted by us
            if not nd.startswith('_') and nd not in nodes:
                util.print_warn('Cannot find matching node "{}" in graph spec'.format(nd))

    def _parse_graph(self):
        graph_spec = self.spec['graph'] if 'graph' in self.spec else []

        try:
            nodes, edges = GraphParser(graph_spec).parse()
            self._match_nodes(nodes)
            self.paths = GraphAnalyzer(nodes, edges).analyze()

            # an ugly way to handle the artificial _start node
            if '_start' in self.blocks:
                for p in self.paths:
                    p.insert(0, '_start')
                if len(self.paths) == 0:
                    self.paths = [['_start']]
        except ParseError as e:
            self._throw_spec_error(e.args[0])
        except InvalidGraphError as e:
            self._throw_spec_error(e.args[0])

    def _get_code_paths(self):
        res = []

        for path in self.paths:
            pt = []
            for nd in path:
                pt.extend(self.blocks[nd].chunks)
            res.append(pt)

        return res

    def _code_gen_recur(self, path, i, code, history):
        if i >= len(path):
            self.counter += 1
            fn = 'universe_{}{}'.format(self.counter, self.lang.get_ext())

            # record history
            history.filename = fn
            self.history.append(history)

            # write file
            fn = os.path.join(self.out, 'code/', fn)
            with open(fn, 'w') as f:
                f.write(code)
                f.flush()
        else:
            val, template = path[i]

            if val != '':
                # check if we have already encountered the decision
                prev_idx = None
                for d in history.decisions:
                    if d[0] == val:
                        prev_idx = d[2]

                if prev_idx is not None:
                    # use the previous value
                    snippet, opt = self.dec_parser.gen_code(template, val, prev_idx)
                    self._code_gen_recur(path, i+1, code+snippet, history)
                else:
                    # expand the decision
                    num_alt = self.dec_parser.get_num_alt(val)
                    for k in range(num_alt):
                        snippet, opt = self.dec_parser.gen_code(template, val, k)
                        decs = [a for a in history.decisions]
                        decs.append((val, opt, k))
                        self._code_gen_recur(path, i+1, code + snippet,
                                             History(history.path, '', decs))
            else:
                code += template
                self._code_gen_recur(path, i+1, code, history)

    def _code_gen(self):
        dir_script = 'code/'

        if os.path.exists(self.out):
            shutil.rmtree(self.out)
        os.makedirs(self.out)
        os.makedirs(os.path.join(self.out, dir_script))

        paths = self._get_code_paths()

        self.counter = 0    # keep track of file name
        self.history = []   # keep track of choices made for each file
        for idx, p in enumerate(paths):
            self._code_gen_recur(p, 0, '', History(idx))

        # output a script to execute all universes
        cmd = self.lang.get_cmd()
        sh = exec_template.format('./{}universe_'.format(dir_script),
                                  self.lang.get_ext(), self.counter, cmd, cmd)
        fn_exec = os.path.join(self.out, 'execute.sh')
        with open(fn_exec, 'w') as f:
            f.write(sh)
        st = os.stat(fn_exec)
        os.chmod(fn_exec, st.st_mode | 0o0111)

    def _write_csv(self):
        with open(os.path.join(self.out, 'summary.csv'), 'w', newline='') as f:
            wrt = csv.writer(f)
            decs = [i for i in self.dec_parser.decisions.keys()]
            wrt.writerow(['Filename', 'Code Path'] + decs)
            for h in self.history:
                row = [h.filename, '->'.join(self.paths[h.path])]
                mp = {}
                for d in h.decisions:
                    mp[d[0]] = d[1]
                for d in decs:
                    value = mp[d] if d in mp else ''
                    row.append(value)
                wrt.writerow(row)

    def _print_summary(self):
        w = 80
        max_rows = 10

        print('=' * w)
        print('{:<20}{:<30}{:<30}'.format('Filename', 'Code Path', 'Decisions'))
        print('=' * w)
        for idx, h in enumerate(self.history):
            path = wrap('->'.join(self.paths[h.path]), width=27)
            decs = ['{}={}'.format(d[0], d[1]) for d in h.decisions]
            decs = wrap(', '.join(decs), width=30)
            max_len = max(len(decs), len(path))

            for r in range(max_len):
                c1 = h.filename if r == 0 else ''
                c2 = path[r] if r < len(path) else ''
                c3 = decs[r] if r < len(decs) else ''

                print('{:<20}'.format(c1), end='')
                print('{:<30}'.format(c2), end='')
                print('{:<30}'.format(c3))
            print('-' * w)

            if idx >= max_rows - 1:
                print('... {} more rows'.format(len(self.history) - max_rows))
                break

    def main(self, verbose=True):
        self._parse_blocks()
        self._parse_graph()
        self._code_gen()
        self._write_csv()
        if verbose:
            self._print_summary()



# -*- coding: utf-8 -*-

"""Console script."""
import click
import os
from .parser import Parser


@click.command()
@click.option('--script', '-s', help='Path to template script',
              default='./script_annotated.py', show_default=True)
@click.option('--json', '-j', help='Path to JSON spec',
              default='./spec.json', show_default=True)
@click.option('--out', '-o', help='Output directory',
              default='.', show_default=True)
@click.option('--lang', '-l', help='Language, can be python/R.',
              default='')
def main(script, json, out, lang):
    """Generate multiverse analysis from specifications."""
    check_path(script)
    check_path(json)

    click.echo('Creating multiverse from {} and {}'.format(script, json))
    ps = Parser(script, json, out, lang)
    ps.main()

    ex = """To execute the multiverse, run the following commands:
    cd {}
    sh execute.sh
    """.format(os.path.join(out, 'multiverse'))
    click.secho('Success!', fg='green')
    click.secho(ex, fg='green')


def check_path(p):
    """Check if the path exists"""
    if not os.path.exists(p):
        msg = 'Error: Path "{}" does not exist.'.format(p)
        print_help(msg)


def print_help(err=''):
    """Show help message and exit."""
    ctx = click.get_current_context()
    click.echo(ctx.get_help())

    if err:
        click.echo('\n' + err)
    ctx.exit()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Ugly hack to allow import from the root folder
import sys
import os
import shutil
sys.path.insert(0, os.path.abspath('..'))

import unittest
from unittest.mock import patch
import io
from boba.parser import Parser

FIRST_SCRIPT = 'multiverse/code/universe_1.py'


def _print_code(ps):
    for b in ps.blocks:
        bl = ps.blocks[b]
        print('-' * 10, '  ({}) {}  '.format(bl.id, bl.name), '-' * 10)
        print(bl.code)


def abs_path(rel_path):
    return os.path.join(os.path.dirname(__file__), rel_path)


class TestParser(unittest.TestCase):

    # --- code gen ---
    # a simple synthetic example
    def test_code_gen(self):
        base = abs_path('../example/simple/')
        ps = Parser(base+'script_annotated.py', base+'spec.json', base)
        ps.main(verbose=False)
        self.assertEqual(ps.counter, 6)

    # a complex example
    def test_codegen_reading(self):
        base = abs_path('../example/reading/')
        Parser(base+'script_annotated.py', base+'spec.json', base).main(verbose=False)

    # another complex example
    def test_codegen_fertility(self):
        base = abs_path('../example/fertility/')
        ps = Parser(base+'script_annotated.py', base+'spec.json', base)
        ps.main(verbose=False)
        self.assertEqual(ps.counter, 120)

    def test_r(self):
        base = abs_path('../example/fertility_r/')
        ps = Parser(base+'template.R', base+'spec.json', base)
        ps.main(verbose=False)

    # the spec has one decision and no graphs; should work
    def test_codegen_decision_only(self):
        base = abs_path('./specs/')
        sc = 'script_no_graph.py'
        ps = Parser(base+sc, base+'spec-no-graph.json', base)
        ps.main(verbose=False)

        # compare file content
        with open(base+FIRST_SCRIPT, 'r') as f:
            actual = f.read()
        with open(base+sc, 'r') as f:
            expected = f.read().replace('{{a}}', '1')
        self.assertEqual(actual, expected)

        # clean up
        shutil.rmtree(base+'multiverse/')

    # the spec has a graph and no decisions; should work
    def test_codegen_graph_only(self):
        base = abs_path('./specs/')
        sc = 'script1.py'
        pout = abs_path('./output_no_dec')
        ps = Parser(base+sc, base+'spec-good.json', pout)
        ps.main(verbose=False)

        # compare file content
        with open(os.path.join(pout, FIRST_SCRIPT), 'r') as f:
            actual = f.read()
        with open(base+sc, 'r') as f:
            lines = f.read().split('\n')
            expected = ''
            for l in lines:
                if not l.strip().startswith('# --- '):
                    expected += l + '\n'
        self.assertEqual(actual.strip(), expected.strip())

        # clean up
        shutil.rmtree(pout)

    # the template contains decisions, but spec does not have corresponding def
    @patch('sys.stdout', new_callable=io.StringIO)
    def test_codegen_missing_decision(self, stdout):
        base = abs_path('./specs/')
        sc = 'script_no_graph.py'
        ps = Parser(base+sc, base+'spec-empty.json', base)
        with self.assertRaises(SystemExit):
            ps.main(verbose=False)
        self.assertRegex(stdout.getvalue(), 'Cannot find the matching variable')

    # --- parse blocks ---
    def test_parse_blocks(self):
        base = abs_path('../example/simple/')
        ps = Parser(base+'script_annotated.py', base+'spec.json')
        ps._parse_blocks()
        self.assertListEqual([*ps.blocks], ['_start', 'A1', 'A2', 'B'])

    def test_script_1(self):
        base = abs_path('./specs/')
        ps = Parser(base+'script1.py', base+'spec-good.json')
        ps._parse_blocks()
        self.assertListEqual([*ps.blocks], ['a', 'b', 'c'])

    def test_script_2(self):
        base = abs_path('./specs/')
        ps = Parser(base+'script2.py', base+'spec-good.json')
        ps._parse_blocks()
        self.assertListEqual([*ps.blocks], ['_start', 'a', 'b', 'c'])

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_script_3(self, stdout):
        base = abs_path('./specs/')
        ps = Parser(base+'script3.py', base+'spec-good.json')
        with self.assertRaises(SystemExit):
            ps._parse_blocks()
        self.assertRegex(stdout.getvalue(), r'Cannot find "\("')

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_script_4(self, stdout):
        base = abs_path('./specs/')
        ps = Parser(base+'script4.py', base+'spec-good.json')
        with self.assertRaises(SystemExit):
            ps._parse_blocks()
        self.assertRegex(stdout.getvalue(), '(?i)duplicated')

    # --- parse graph ---
    def test_spec_good(self):
        base = abs_path('../example/simple/')
        ps = Parser(base+'script_annotated.py', base+'spec.json')
        ps._parse_blocks()
        ps._parse_graph()
        expected = [['_start', 'A1', 'B'], ['_start', 'A2', 'B']]
        expected = set([','.join(p) for p in expected])
        self.assertSetEqual(set([','.join(p) for p in ps.paths]), expected)

    def test_spec_empty(self):
        base = abs_path('./specs/')
        ps = Parser(base+'script1.py', base+'spec-empty.json')
        ps._parse_graph()
        self.assertListEqual(ps.paths, [])

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_spec_bad_graph(self, stdout):
        base = abs_path('./specs/')
        ps = Parser(base+'script1.py', base+'spec-bad-graph.json')
        with self.assertRaises(SystemExit):
            ps._parse_graph()
        self.assertRegex(stdout.getvalue(), 'Cannot find a target node')

    @patch('sys.stdout', new_callable=io.StringIO)
    def test_spec_cyclic_graph(self, stdout):
        base = abs_path('./specs/')
        ps = Parser(base + 'script1.py', base + 'spec-cyclic-graph.json')
        with self.assertRaises(SystemExit):
            ps._parse_blocks()
            ps._parse_graph()
        self.assertRegex(stdout.getvalue(), 'Cannot find any starting node')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

from .baseparser import BaseParser, ParseError

kw = '# ---'


class BlockParser(BaseParser):
    """
    Parse the metadata of a code block, which must have the structure:
        # --- (ID) name of the block
    """

    def __init__(self, line):
        super(BlockParser, self).__init__(line)

        self.state = 0
        self.parsed_id = ''
        self.parsed_name = ''

    @staticmethod
    def can_parse(line):
        return line.lstrip().startswith(kw)

    def parse(self):
        while not self._is_end():
            self._read_next()
        return self.parsed_id, self.parsed_name

    def _read_next(self):
        self._read_while(BlockParser._is_whitespace)
        if self._is_end():
            return

        if self.state == 0:
            self._read_kw()
        elif self.state == 1:
            self._read_id()
        else:
            self._read_name()

    def _end(self):
        self.i = len(self.line)  # stop parsing

    def _throw(self, msg):
        err = 'At character {} of "{}":\n\t{}'.format(self.i+1, self.line, msg)
        raise ParseError(err)

    def _remaining(self):
        return self.line[self.i:]

    def _read_kw(self):
        if self._remaining().startswith(kw):
            self.i += len(kw)
            self.state += 1
        else:
            self._throw('expected {}'.format(kw))

    def _read_id(self):
        # open paren
        if self._peek_char() != '(':
            self._throw('Cannot find "("')
        self._next_char()

        # read the actual identifier
        ch = self._peek_char()
        if not self._is_id_start(ch):
            self._throw('Invalid identifier start character {}'.format(ch))

        self.parsed_id = self._read_while(self._is_id)

        # close paren
        if self._peek_char() != ')':
            self._throw('Cannot find ")"')
        self._next_char()
        self.state += 1

    def _read_name(self):
        self.parsed_name = self._remaining().strip()
        self._end()



#!/usr/bin/env python3

from dataclasses import dataclass
from .baseparser import BaseParser, Token, ParseError


@dataclass(frozen=True)
class Edge:
    start: str
    end: str


class GraphParser(BaseParser):
    def __init__(self, graph_spec):
        line = '\n'.join(graph_spec)
        super(GraphParser, self).__init__(line)

        self.spec = graph_spec
        self.nodes = set()
        self.edges = set()

    def _prep_err(self, msg):
        return 'At character {} of "{}":\n\t{}'.format(self.col+1, self.spec[self.row], msg)

    def _read_next(self):
        self._read_while(GraphParser._is_whitespace)
        if self._is_end():
            return Token('eof', '')

        ch = self._peek_char()
        if ch == '-':
            return self._read_edge()
        elif GraphParser._is_id_start(ch):
            return self._read_node()
        else:
            raise ParseError(self._prep_err('Cannot handle character "{}"'.format(ch)))

    def _read_edge(self):
        val = self._next_char()
        ch = self._peek_char()
        if ch != '>':
            raise ParseError(self._prep_err('Expected "->", got "-{}"'.format(ch)))
        val += self._next_char()
        return Token('edge', val)

    def _read_node(self):
        nd = self._read_while(GraphParser._is_id)
        return Token('node', nd)

    def parse(self):
        prev_node = None

        while True:
            tk = self._next()

            if tk.type == 'node':
                self.nodes.add(tk.value)
                prev_node = tk.value
            if tk.type == 'edge':
                if not prev_node:
                    raise ParseError(self._prep_err('Cannot find a source node'))
                nx = self._peek()
                if nx.type != 'node':
                    raise ParseError(self._prep_err('Cannot find a target node'))
                self.edges.add(Edge(prev_node, nx.value))
            if tk.type == 'eof':
                break

        return self.nodes, self.edges
